#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.5

$LastChangedRevision: 6103 $
$LastChangedDate:: 2022-08-22 16:33:52 +0200#$

This file is part of SyMSpace.

Copyright (c) 2017, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import datetime
import glob
import optparse
import os
import stat
import shutil
import uuid
from os.path import join as pathjoin
import pandas as pd
import xlsxwriter
import componentspace.osutils.fileutils as fu
import componentspace.osutils.pkgutils as pkgutils
from componentspace.create_buildscript import ComponentBuilder
from componentspace.report.json import JSONReport
from componentspace.osutils.paths import ComponentPaths
from componentspace.base_component import Component

from symspace.osutils.pathutils import Paths
from symspace.osutils.pathutils import CMDParser
from symspace.wrapper.symwrapper import SymProject, SymContainer, SymFile, SymField


class ComponentManager:

    def __init__(self, start_path: str):
        """
        Component manager class

        Parameters
        ----------
        start_path : str
            Path to the directory where build starts.
            Components in all subdirectories will be built.
        """
        self.start_path = start_path
        self.json_file = os.path.join(self.start_path, 'component_info.json')
        self.xls_file = os.path.join(self.start_path, 'component_info.xlsx')

    @classmethod
    def fromCMDParser(cls) -> 'ComponentManager':
        """
        Commandline parser to create instance of this class

        Commandline Arguments
        ---------------------
        --startDir : str
            The search for component starts from this directory (Optional).
            Default is the ComponentSpace root directory.

        Returns
        -------
        manager : ComponentManager
        """

        parser = optparse.OptionParser("usage: %prog [options]")
        parser.add_option("-d", "--startDir", dest="start_dir", default=Paths.getComponentSpaceRoot(),
                          type="string", help="The search for components starts from this directory, "
                                              "default is the ComponentSpace root directory.")

        args = CMDParser.jepSysArgs()
        (options, args) = parser.parse_args(args)
        if len(args) != 0:
            parser.error("incorrect number of arguments")

        start_dir = CMDParser.jepArgPath(options.start_dir)

        print('Start Path: {}'.format(start_dir))

        manager = cls(start_dir)
        return manager

    def collectInfo(self, components_list=None, all_components=False):
        """
        Component informations are written to a json/excel file
        """
        if components_list is None:
            components_list = self._getComponentsList(all_components)

        report = JSONReport(self.json_file)
        n_components = len(components_list)
        cs_root = Paths.getComponentSpaceRoot()
        for index, path_comp in enumerate(components_list):
            disp_path = path_comp.replace(cs_root, '.')
            tmp_webpath = os.path.dirname(path_comp).replace(cs_root, '').replace('\\', '.')[1:]
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Collect information {}/{}: {}'.format(timestamp, index + 1, n_components, disp_path))

            try:
                status, prj = self._getProject(path_comp)

                usergroup = prj.getObject('Filter.UserGroup').getData()
                webpath = prj.getObject('Info.WebPath').getData()
                version = prj.getObject('Info.Version').getData()
                pkgs = pkgutils.findUsedPackages(prj, get_version=True)

                report.build[path_comp] = {'UserGroup': usergroup, 'WebPath': webpath, 'Provides': [], 'Requires': [], 'Status': status}
                if webpath != tmp_webpath:
                    report.build[path_comp]['Webpath (Deduced from directory structure)'] = tmp_webpath
                report.build[path_comp]['Version'] = version
                report.build[path_comp]['Packages'] = pkgs

                param_provides = prj.getObject('Filter.provides')
                if param_provides is not None:
                    if isinstance(param_provides, SymContainer):
                        for par in param_provides.getParameters():
                            report.build[path_comp]['Provides'].append({'Name': par.getName(),
                                                                        'Desc': par.getDescription()})

                param_requires = prj.getObject('Filter.requires')
                if param_requires is not None:
                    if isinstance(param_requires, SymContainer):
                        for par in param_requires.getParameters():
                            report.build[path_comp]['Requires'].append({'Name': par.getName(),
                                                                        'Desc': par.getDescription()})

                '''HINT: if closing the project results in JRE ACCESS VIOLATION ERROR due to python*.dll 
                search for the responsible python package and add in the config at JepSharedModules'''
                prj.close()

            except Exception as e:
                report.build[path_comp] = {}
                print('\tError: {}'.format(str(e)))

        '''write files at the end'''
        report.writeJSON()
        writer = pd.ExcelWriter(self.xls_file, engine='xlsxwriter')
        print('\nwriting "All Components"')
        report.appendXLS(writer, 'All Components')
        print('\nwriting "version_subzero" components')
        report.appendXLS(writer, 'version_subzero', "Version", "1.0", "<")
        print('\nwriting "disabled" components')
        report.appendXLS(writer, 'disabled', "Status", "disabled", "=")
        print('\nwriting "deprecated" components')
        report.appendXLS(writer, 'deprecated', "Status", "deprecated", "=")
        writer.save()
        writer.close()

        print('\nFile saved to: {}'.format(self.json_file))
        print('\nFile saved to: {}'.format(self.xls_file))

    def refactorField(self, identifier: str,
                      name: str = None, description: str = None, comment: str = None,
                      unit: str = None, tex_string: str = None, type_string: str = None, dry_run: bool = False,
                      components_list=None):
        """
        Try refactoring a field of all found components if the field exists

        Change: e.g. Name, Description, Comment, Unit, TexString, Type

        Parameters
        ----------

        identifier: str
            Identifier of the field
        name: str
            New name, if not None (Default)
        description: str
            New description, if not None (Default)
        comment: str
            New comment, if not None (Default)
        unit: str
            New unit, if not None (Default)
        tex_string: str
            New tex_string, if not None (Default)
        type_string: str
            New type_string, if not None (Default)
        dry_run: bool
            Do not save the changes, just display them
        components_list
        """
        if components_list is None:
            components_list = self._getComponentsList()

        n_components = len(components_list)
        cs_root = Paths.getComponentSpaceRoot()
        for index, path_comp in enumerate(components_list):
            disp_path = path_comp.replace(cs_root, '.')
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Refactoring {}/{}: {}'.format(timestamp, index + 1, n_components, disp_path))

            try:
                status, prj = self._getProject(path_comp)
                param = prj.getObject(identifier)
                if isinstance(param, SymField):
                    changed = False
                    if name is not None:
                        old_name = param.getName()
                        if old_name != name:
                            param.setName(name)
                            print('\tChanging Name {} => {}'.format(old_name, name))
                            changed = True
                    if description is not None:
                        old_description = param.getDescription()
                        if old_description != description:
                            param.setDescription(description)
                            print('\tChanging Description {} => {}'.format(old_description, description))
                            changed = True
                    if comment is not None:
                        old_comment = param.getComment()
                        if old_comment != comment:
                            param.setComment(comment)
                            print('\tChanging Comment {} => {}'.format(old_comment, comment))
                        changed = True
                    if unit is not None:
                        old_unit = param.getUnit()
                        if old_unit != unit:
                            param.setUnit(unit)
                            print('\tChanging Unit {} => {}'.format(old_unit, unit))
                            changed = True
                    if tex_string is not None:
                        old_tex_string = param.getTexString()
                        if old_tex_string != tex_string:
                            param.setTexString(tex_string)
                            print('\tChanging TexString {} => {}'.format(old_tex_string, tex_string))
                            changed = True
                    if type_string is not None:
                        old_type_string = param.getType()
                        if old_type_string != type_string:
                            param.setType(type_string)
                            print('\tChanging Type {} => {}'.format(old_type_string, type_string))
                            changed = True
                    if changed and not dry_run:
                        prj.save()

                else:
                    print('\tWarning: No field {} found'.format(identifier))

                prj.close()

            except Exception as e:
                print('\tError: {}'.format(str(e)))

    def addPackages(self, packages: tuple, auto_collect: bool, components_list=None):
        """
        Add new packages

        Parameters
        ----------
        packages : tuple
        auto_collect:bool
        components_list
        """
        if components_list is None:
            components_list = self._getComponentsList()

        n_components = len(components_list)
        cs_root = Paths.getComponentSpaceRoot()
        for index, path_comp in enumerate(components_list):
            disp_path = path_comp.replace(cs_root, '.')
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Add Packages {}/{}: {}'.format(timestamp, index + 1, n_components, disp_path))

            try:
                changed_flag = False
                status, prj = self._getProject(path_comp)
                parent = prj.getObject('Packages')
                if parent is None:
                    '''add Packages container if not already existing'''
                    prj.addParam(SymContainer('Packages', 'Packages'))
                    parent = prj.getObject('Packages')  # type: SymContainer

                '''collect existing packages'''
                existing_packages = []
                for param in parent.getParam():
                    if isinstance(param, SymField):
                        pkg_name = param.getData()
                        pkg_name = pkg_name.split('.')[0]
                        existing_packages.append(pkg_name)
                        # if param.getName() != pkg_name:
                        #     '''rename param to package name'''
                        #     param.setName(pkg_name)

                '''remove duplicates'''
                existing_packages = set(existing_packages)

                if auto_collect:
                    _, copy_pkgs = Component.collectPythonDependencies(prj, fnc_files=None, auto_collect=True)
                    tmp_packages = list(copy_pkgs.keys())
                    tmp_packages.extend(packages)
                    tmp_packages = set(tmp_packages)
                else:
                    tmp_packages = set(packages)

                '''add new packages'''
                for pkg_name in tmp_packages:
                    if pkg_name not in existing_packages:
                        param = SymField(pkg_name)
                        parent.addParam(param)
                        param.setData(pkg_name)
                        changed_flag = True
                        print('\t \\--> added new package {}'.format(pkg_name))
                    else:
                        print('\t \\--> package is already included {}'.format(pkg_name))

                if changed_flag:
                    '''only save if changes were done'''
                    prj.save()
                prj.close()

            except Exception as e:
                print('\tError: {}'.format(str(e)))

    def refactorPackages(self, package_replacement: dict, remove_subpkg: bool = False, components_list=None):
        """
        Replace old package names with new names

        Parameters
        ----------
        package_replacement : dict
            key: old pkg name
            value: new pkg name
            lcm_motorbox.geometry > lcm_geometry
        remove_subpkg : bool
            replace the package path with the toplevel package
            e.g. lcm_motorbox.postprocessing.xy -> lcm_motorbox
        components_list
        """
        if components_list is None:
            components_list = self._getComponentsList()

        n_components = len(components_list)
        cs_root = Paths.getComponentSpaceRoot()
        for index, path_comp in enumerate(components_list):
            disp_path = path_comp.replace(cs_root, '.')
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Refactoring {}/{}: {}'.format(timestamp, index + 1, n_components, disp_path))

            try:
                status, prj = self._getProject(path_comp)
                changed_flag = False
                if prj.getObject('Packages') is not None:
                    for param in prj.getObject('Packages').getParam():
                        if isinstance(param, SymField):
                            pkg_name = param.getData()
                            pkg_path = pkg_name.split('.')
                            new_pkg_name = None

                            if remove_subpkg and len(pkg_path) > 1:
                                new_pkg_name = pkg_path[0]

                            if pkg_name in package_replacement:
                                new_pkg_name = package_replacement[pkg_name]

                            if new_pkg_name is not None:
                                param.setData(new_pkg_name)
                                changed_flag = True
                                print('\t \\--> replaced package {} with {}'.format(pkg_name, new_pkg_name))

                if changed_flag:
                    '''only save if changes were done'''
                    prj.save()
                prj.close()

            except Exception as e:
                print('\tError: {}'.format(str(e)))

    def changeVersion(self, major: int, minor: int, increase_version: bool, components_list=None):
        """
        Change the component version

        Parameters
        ----------
        components_list: list
        major: int
            Major version
        minor: int
            Minor version
        increase_version : bool
            If True: Increase old version by major.minor
            Else: Replace old version by major.minor
        components_list
        """

        if components_list is None:
            components_list = self._getComponentsList()

        n_components = len(components_list)
        cs_root = Paths.getComponentSpaceRoot()
        for index, path_comp in enumerate(components_list):
            disp_path = path_comp.replace(cs_root, '.')
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Changing version {}/{}: {}'.format(timestamp, index + 1, n_components, disp_path))

            try:
                status, prj = self._getProject(path_comp)
                param = prj.getObject('Info.Version')
                if param is not None:
                    old_version = param.getData()

                    if increase_version:
                        old_major, old_minor = old_version.split('.')
                        new_major = str(int(old_major) + major)
                        new_minor = str(int(old_minor) + minor)
                    else:
                        new_major = str(major)
                        new_minor = str(minor)

                    new_version = '.'.join([new_major, new_minor])

                    param.setData(new_version)
                    prj.save()
                    print('\t \\--> updated version {} => {}'.format(old_version, new_version))

                    prj.close()

            except Exception as e:
                print('\tError: {}'.format(str(e)))

    @staticmethod
    def _getProject(path_comp: str) -> SymProject:
        files = glob.glob(os.path.join(path_comp, '*.mop'))
        if len(files) > 1:
            raise Exception('Found multiple mop-files in the folder')
        elif len(files) == 0:
            raise Exception('Found no mop-files in the folder')

        build_file = glob.glob(os.path.join(path_comp, 'build_component.py'))
        disabled_file = glob.glob(os.path.join(path_comp, 'disabled_build_component.py'))
        deprecated_file = glob.glob(os.path.join(path_comp, 'deprecated_build_component.py'))
        if len(build_file) + len(disabled_file) + len(deprecated_file) != 1:
            raise Exception('Found multiple or none build_scripts')
        else:
            if len(build_file) > 0:
                status = 'enabled'
            elif len(disabled_file) > 0:
                status = 'disabled'
            else:
                status = 'deprecated'

        prj = SymProject.load(files[0])
        return status, prj

    def _getComponentsList(self, all_components=False):
        builder = ComponentBuilder(self.start_path, full_build=True)
        components_list = builder.getChangedComponents(write_xml=False, all_components=all_components)
        return components_list


def moveComponent(src, dst):
    """
    Move Component to different location

    Parameters
    ----------
    src : str
        Path to Component directory
    dst: str
        Path to Component final directory
    """
    fu.copyTree(src, dst)
    relsrc = src.replace(Paths.getComponentSpaceRoot(), '')
    reldst = dst.replace(Paths.getComponentSpaceRoot(), '')
    fu.replaceInFile(pathjoin(dst, 'build_component.py'), relsrc.replace('\\', '\\\\'), reldst.replace('\\', '\\\\'))
    # print(pathjoin(dst, reldst.split('\\')[-1]+'.mop'))
    sym_mdl = SymProject.load(pathjoin(dst, reldst.split('\\')[-1] + '.mop'))
    sym_mdl['WebPath'].setData('.'.join(reldst.split('\\')[0:-1]))
    sym_mdl.save()


def getComponents(start_path):
    """ Get all Components within start path"""
    lst_comp = []
    for fh in fu.searchBuildComponentScripts(start_path):
        path = os.path.dirname(fh)
        name = os.path.split(path)[-1] + '.mop'
        fname = pathjoin(path, name)
        if os.path.isfile(fname):
            lst_comp.append(fname)
    return lst_comp


def getTemplates(start_path):
    """ Get template Components within start path"""
    lst_comp = []
    for fname in glob.iglob(start_path + '/**/Template.mop', recursive=True):
        if '_template' in fname:
            lst_comp.append(fname)
    return lst_comp


def removeBuildInfo(start_path):
    for fh in glob.iglob(start_path + '/**/' + ComponentPaths.build_file, recursive=True):
        print('Remove: {}'.format(fh))
        os.remove(fh)
    for fh in glob.iglob(start_path + '/**/' + ComponentPaths.revision_file, recursive=True):
        print('Remove: {}'.format(fh))
        os.remove(fh)


def exportComponentSpace(location):
    remove_files = [
        '_build',
        'Manufacture',
        'KnowledgeEngines',
        '_UserGroups\\BR',
        '_UserGroups\\ECS',
        '_UserGroups\\Hanning',
        '_UserGroups\\LCM_Drives',
        '_UserGroups\\LTI',
        '_UserGroups\\MotorBox',
        '_UserGroups\\Piezo',
        '_UserGroups\\Vibration',
        'src\\tests',
        'src\\buildscripts',
        'src\\componentspace\\.idea',
        'src\\componentspace\\playground',
        'Simulation\\Motor\\Magnetic\\PMSM\\PMSM_Template',
        'Simulation\\Motor\\Magnetic\\PMSMGereon',
        'Material\\Metal\\MagneticAlloy\\BomatecMagnets'
    ]

    print('Export ComponentSpace to: {}'.format(location))
    src = Paths.getComponentSpaceRoot()
    if os.path.exists(location):
        shutil.rmtree(location, onerror=fu.on_rm_error)

    if not os.path.exists(location):
        os.mkdir(location)

    fu.copyTree(src, location)

    # Remove Components
    for comp in getComponents(location):
        comp_path = os.path.dirname(comp)
        print('Remove: {}'.format(comp_path))
        if os.path.exists(comp_path):
            shutil.rmtree(comp_path, onerror=fu.on_rm_error)

    # Remove Templates
    for comp in getTemplates(location):
        comp_path = os.path.dirname(comp)
        if comp_path == os.path.join(location, '_template'):
            '''keep base template'''
            continue
        print('Remove: {}'.format(comp_path))
        if os.path.exists(comp_path):
            shutil.rmtree(comp_path, onerror=fu.on_rm_error)

    # Remove build information
    removeBuildInfo(location)

    # Empty UserGroups
    print('Emptying _UserGroups ...')
    for gr in os.listdir(pathjoin(location, '_UserGroups')):
        path_comp = pathjoin(location, '_UserGroups', gr, 'Components')
        if os.path.exists(path_comp):
            for fh in os.listdir(path_comp):
                print('Remove ' + fh)
                file = pathjoin(path_comp, fh)
                if os.path.isfile(file):
                    os.remove(file)
                elif os.path.exists(file):
                    shutil.rmtree(file)

    # Remove files and directories from list
    for item in remove_files:
        fh = os.path.join(location, item)
        if os.path.isdir(fh):
            print('Delete directory: {}'.format(fh))
            shutil.rmtree(fh, onerror=fu.on_rm_error)
        elif os.path.isfile(fh):
            print('Delete file: {}'.format(fh))
            os.remove(fh)
        else:
            print('Cannot delete: {}'.format(fh))

    # Empty packages
    for item in os.listdir(os.path.join(location, 'src', 'packages')):
        fh = os.path.join(location, 'src', 'packages', item)
        if os.path.isdir(fh):
            print('Delete directory: {}'.format(fh))
            shutil.rmtree(fh, onerror=fu.on_rm_error)


def exportParameter(start_path, writeall=True):
    # Create an new Excel file and add a worksheet.
    workbook = xlsxwriter.Workbook(os.path.join(start_path, 'Parameter.xlsx'))
    worksheet = workbook.add_worksheet()

    line = 1

    worksheet.write(0, 0, 'Component')
    worksheet.write(0, 1, 'Identifier')
    worksheet.write(0, 2, 'Name')
    worksheet.write(0, 3, 'Description')
    worksheet.write(0, 4, 'Tex string')
    worksheet.write(0, 5, 'Unit')

    parlst = []
    components = getTemplates(start_path)
    components.extend(getComponents(start_path))
    for comp in components:
        prj = SymProject.load(comp)
        for par in prj['Project'].getParameters('SymParameter'):
            if not isinstance(par, (SymContainer, SymFile)):
                lst = (par.getName(), par.getTexString(), par.getUnit())
                if writeall or (lst not in parlst):
                    parlst.append(lst)
                    worksheet.write(line, 0, comp)
                    worksheet.write(line, 1, par.getIdentifier())
                    worksheet.write(line, 2, par.getName())
                    worksheet.write(line, 3, par.getDescription())
                    worksheet.write(line, 4, par.getTexString())
                    worksheet.write(line, 5, par.getUnit())
                    line += 1
                    print(
                        '{}, {}, {}, {}'.format(par.getName(), par.getDescription(), par.getTexString(), par.getUnit()))

    workbook.close()


def updateParameter():
    """
    Standardisation of Parameter Name, Description, TexString and Unit

    Achtung: Beim Sortieren des Excel Files ist mir ein Fehler passiert. Component und der Rest passen nicht zusammen,
    daher ist der Code so umständlich

    """
    xls_file = os.path.join(os.getenv('ComponentSpace'), 'Simulation', 'Motor', 'Magnetic', 'PMSM', 'rst',
                            'Parameter.xlsx')
    df = pd.read_excel(xls_file)
    df.fillna('', inplace=True)

    components = df['Component'].unique()

    for comp in components:
        prj = SymProject.load(comp)
        update = False

        for par in prj['Project'].getParameters('SymParameter'):
            if not isinstance(par, (SymContainer, SymFile)):
                ident = par.getIdentifier()
                res = df[df['Identifier'] == ident]
                name = res['Name'].unique()
                descr = res['Description'].unique()
                tex = res['Tex string'].unique()
                unit = res['Unit'].unique()

                if len(name) == 1 and len(descr) == 1 and len(tex) == 1 and len(unit) == 1:

                    if name[0] != par.getName():
                        update = True
                        par.setName(name[0])
                    if descr[0] != par.getDescription():
                        update = True
                        par.setDescription(descr[0])
                    if tex[0] != par.getTexString():
                        update = True
                        par.setTexString(tex[0])
                    if unit[0] != par.getUnit():
                        update = True
                        par.setUnit(unit[0])

                    if update:
                        print('{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}'.format(name[0], par.getName(), descr[0],
                                                                      par.getDescription(),
                                                                      tex[0], par.getTexString(), unit[0],
                                                                      par.getUnit()))

                else:
                    print('********** Comp: {}, \tIdent: {} not unique!'.format(comp, ident))

        if update:
            prj.save()
        prj.close()


def addStepCreo(start_path):
    for comp in getComponents(start_path):
        prj = SymProject.load(comp)
        if prj['Creo.Add.DRWList.STEP'] is None:
            drwlist = prj['Creo.Add.DRWList']
            if drwlist is not None:
                print(comp)
                drwlist.addParam(prj['Creo.Add.Assembly'].cloneParam())
                prj['Creo.Add.DRWList.Assembly'].setName('STEP')
                prj.save()


def updateFilter(start_path):
    """
    Replace SimData in Filter property by provides_type

    Parameters
    ----------
    start_path : str
        Path to change Components

    """
    components = getTemplates(start_path)
    components.extend(getComponents(start_path))
    cnt = 0
    for comp in components:
        prj = SymProject.load(comp)
        update = False
        for par in prj['Filter'].getParameters('SymField'):
            if par.getName() == 'property' and par.getData() == 'SimData':
                print('Name: {}, \tComp: {}'.format(par.getIdentifier(), comp))
                par.setData('provides_type')
                update = True

        for par in prj['Filter'].getParameters('SymContainer'):
            if par.getName() == 'Provides':
                print('Name: {}, \tComp: {}'.format(par.getIdentifier(), comp))
                par.removeParam()
                update = True

            if par.getName() == 'Requires':
                print('Name: {}, \tComp: {}'.format(par.getIdentifier(), comp))
                par.removeParam()
                update = True

        if update:
            prj.save()
            cnt += 1

        prj.close()

    print('Number of updated {}'.format(cnt))


def updateFilter1(start_path):
    """
    Replace SimData in Filter property by provides_type

    Parameters
    ----------
    start_path : str
        Path to change Components

    """
    components = getTemplates(start_path)
    components.extend(getComponents(start_path))
    cnt = 0
    for comp in components:
        prj = SymProject.load(comp)
        update = False
        if prj['Filter.provides'] is not None:
            for par in prj['Filter.provides'].getParameters('SymField'):
                if par.getName() == 'property' and par.getData() == 'provides_type':
                    print('Name: {}, \tComp: {}'.format(par.getIdentifier(), comp))
                    par.setData('type')
                    update = True

        if update:
            prj.save()
            cnt += 1

        prj.close()

    print('Number of updated {}'.format(cnt))


def addPackages2Geometry(start_path):
    components = getTemplates(start_path)
    components.extend(getComponents(start_path))
    cnt = 0
    for comp in components:
        prj = SymProject.load(comp)
        update = False
        if prj['Packages'] is None:
            prj.addParam(SymContainer('Packages', 'Python packages to be included in Component'))
            for _ in range(3):
                prj['Packages'].moveUp()
            update = True

        if prj['Packages.Geometry'] is None:
            prj['Packages'].addParam(SymField('Geometry', 'Geometry package'))
            prj['Packages.Geometry'].setData('lcm_motorbox.geometry')
            update = True

        geo = prj['Project'].getParam()[0]

        if geo['Placement'] is not None and geo['Placement'].getUnit() == 'mm':
            geo['Placement'].setUnit('')
            update = True

        if update:
            print('Update Component: {}'.format(comp))
            prj.save()
            cnt += 1

        prj.close()


def getBaseSimulation(start_path):
    for comp in getComponents(start_path):
        prj = SymProject.load(comp)

        if prj['Info.Type'].getData() == 'BaseModel':
            print(comp)

        prj.close()


def setDisplayName(start_path):
    components = getTemplates(start_path)
    components.extend(getComponents(start_path))

    for comp in components:
        prj = SymProject.load(comp)

        if prj['Info.DisplayName'] is None:
            print(comp)
            name = SymField('DisplayName')
            prj['Info'].addParam(name, 2)
            display = prj['Info.Name'].getData().replace('_', ' ')
            name.setData(display)
            prj.save()

            info = os.path.join(os.path.dirname(comp), 'index.rst')
            if os.path.exists(info) and display != prj['Info.Name'].getData():
                content = []
                head = 0
                with open(info, 'r') as fh:
                    for k, line in enumerate(fh.readlines()):
                        if head == 0 and line.startswith('=='):
                            head = k - 1
                        content.append(line)
                content[head] = display + '\n'
                content[head + 1] = '=' * len(display.strip()) + '\n'
                # print(content)

                # Release lock
                os.chmod(info, stat.S_IWRITE)
                with open(info, 'w') as fh:
                    for line in content:
                        fh.write(line)

        prj.close()


def setPlacement(start_path):
    components = getTemplates(start_path)
    components.extend(getComponents(start_path))
    cnt = 0

    for comp in components:
        prj = SymProject.load(comp)

        if not prj['Filter.placement']:
            placement = SymField('placement', 'Position of component in project')
            placement.setComment("Predefined positions: 'parent', 'root', 'self'")
            prj['Filter'].addParam(placement)
            placement.setData('{parent}.Geometry')
            cnt += 1
            prj.save()

    print(f'Modified Components: {cnt}')


def setGUID(start_path):
    components = getComponents(start_path)
    for comp in components:
        print(comp)
        prj = SymProject.load(comp)
        modified = False

        if prj['Filter.provides']:
            for par in prj['Filter.provides'].getParameters('SymField'):
                if par.getName() == 'value':
                    rule = par.getParent()
                    rule_param = [p.getName() for p in rule.getParam()]
                    if 'guid' not in rule_param:
                        guid = SymField('guid', 'Globally Unique Identifier')
                        rule.addParam(guid)
                        guid.setData(str(uuid.uuid4()))
                        modified = True

                    if 'displayName' not in rule_param:
                        print(f'Name {par.getData()}')
                        name = SymField('displayName', 'displayName')
                        rule.addParam(name)
                        name.setData(par.getData().replace('_', ' '))
                        modified = True

        if prj['Filter.requires']:
            for par in prj['Filter.requires'].getParameters('SymField'):
                if par.getName() == 'displayName':
                    rule = par.getParent()
                    rule_param = [p.getName() for p in rule.getParam()]
                    if 'guid' not in rule_param:
                        guid = SymField('guid', 'Globally Unique Identifier')
                        rule.addParam(guid)
                        guid.setData(str(uuid.uuid4()))
                        modified = True

        if modified:
            prj.save()

        prj.close()


def updatePartFilter(start_path):
    components = getComponents(start_path)
    for comp in components:
        print(comp)
        prj = SymProject.load(comp)
        modified = False
        if not prj['Filter.provides']:
            continue
        for rule in prj['Filter.provides'].getParam():
            value = rule['value'].getData()
            if value.startswith('FEM'):
                rule['value'].setData(value.replace('_Creo', '').replace('_FreeCAD', ''))
                modified = True
        if modified:
            prj.save()

        prj.close()


# Execute in CSyMSpace
if __name__ == '__main__':
    updatePartFilter(r'C:\ComponentSpace\Part\MotorPart')
    # setGUID(r'C:\ComponentSpace\Simulation\Motor\Magnetic\PMSM\PMSM_Interior_FEMM_V2')
    # setGUID(r'C:\ComponentSpace\Part')
    # setPlacement(r'C:\ComponentSpace\Part')
    # setDisplayName(Paths.getComponentSpaceRoot())
    # setDisplayName(r'C:\ComponentSpace\Part\MotorPart\StatorInterior\Stator_Interior1')
    # exportComponentSpace(r'C:\Users\ssilber\Documents\ComponentSpaceExport')
    # getBaseSimulation(Paths.getComponentSpaceRoot())

    # addPackages2Geometry(os.path.join(Paths.getComponentSpaceRoot(), 'Part'))
    # exportParameter(r'C:\ComponentSpace\Material')
    print('\nScript finished!')
    # print('Number of Components: {}'.format(len(getComponents('C:\\ComponentSpace'))))
