#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.5

$LastChangedRevision: 6103 $
$LastChangedDate:: 2022-08-22 16:33:52 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import shutil
from os.path import join as pathjoin
import time
import datetime
import optparse
import json
import symspace.osutils.processutils as pu
from symspace.osutils.pathutils import Paths
from symspace.osutils.pathutils import CMDParser
from symspace.pyutils.pkgutil import PkgUtil
from symspace.pyutils.version_util import VersionUtil
import componentspace.osutils.fileutils as fu
from componentspace.help.component_help import ComponentHelp
from componentspace.osutils.paths import ComponentPaths
from componentspace.osutils.svnutils import ComponentSVNUtils
from componentspace.utils.sorting import kahn_topsort
from componentspace.builders.scripts import run_external_buildscript
from componentspace.report.junit import JUnitReport
from componentspace.report.json import JSONReport


class ComponentBuilder:
    build_reports_dir = os.path.join(Paths.getCSTestReportsDir(), 'build')

    def __init__(self, start_path: str, full_build: bool, timeout: float = 600):
        """
        Component builder class

        Parameters
        ----------

        start_path : str
            Path to the directory where build starts.
            Components in all subdirectories will be built.
        full_build : bool
            If True all components are built
        timeout : float
            Timeout (in seconds) for single component build script
        """
        self.start_path = start_path
        self.full_build = full_build
        self.allow_include_missing = False
        self.stable_build = False

        '''timeout for component build subprocess in seconds'''
        self.timeout = timeout

        '''JSON file with last versions of last build'''
        self.build_info = pathjoin(start_path, 'build_info.json')

        '''report files'''
        changed_results_txt = pathjoin(start_path, 'changed_results.txt')
        build_results_txt = pathjoin(start_path, 'build_results.txt')

        ''' unittest files'''
        # timestr = time.strftime("%Y%m%d-%H%M%S")
        changed_results_xml = pathjoin(start_path, 'changed_results-junit.xml')
        build_results_xml = pathjoin(start_path, 'build_results-junit.xml')

        self.json_report = JSONReport(self.build_info)
        self.junit_changed = JUnitReport('componentspace.create_buildscript.Changed', changed_results_xml,
                                         changed_results_txt)
        self.junit_build = JUnitReport('componentspace.create_buildscript.Build', build_results_xml, build_results_txt)

        pkg_paths = PkgUtil.findSrcPackages([os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages'), ])

        self.pkg_versions = dict()
        for pkg_name, pkg_path in pkg_paths.items():
            pkg_version = PkgUtil.getSrcPackageVersion(pkg_name, pkg_path)
            if pkg_version is not None:
                self.pkg_versions[pkg_name] = pkg_version

    @classmethod
    def fromCMDParser(cls) -> 'ComponentBuilder':
        """
        Commandline parser to create instance of this class

        Commandline Arguments
        ---------------------

        --fullBuild : int
            If --fullBuild is 1 all components will be built,
            if --fullBuild is 0 (Default) only new components will be built.

        --startDir : str
            The search for component starts from this directory (Optional).
            Default is the ComponentSpace root directory.

        --timeout : float
            Timeout (in milliseconds) for single component build script

        Returns
        -------
        builder : ComponentBuilder
        """

        parser = optparse.OptionParser("usage: %prog [options]")
        parser.add_option("-f", "--fullBuild", dest="full_build",
                          default=0, type="int",
                          help="(re)-build all components, default is 0 (False)")
        parser.add_option("--buildType", dest="build_type",
                          default='nightly', type="str",
                          help="(re)-build only stable components, default is 0 (False)")
        parser.add_option("-d", "--startDir", dest="start_dir", default=Paths.getComponentSpaceRoot(),
                          type="string", help="The search for components starts from this directory, "
                                              "default is the ComponentSpace root directory.")
        parser.add_option("-t", "--timeout", dest="timeout", default=600,
                          type="float", help="Timeout (in seconds) for single component build script")
        parser.add_option("--IncludeMissingDeps", dest="allow_include_missing",
                          default=0, type="int",
                          help="Allow to include missing dependencies to the component build list, "
                               "default is 0 (False)")

        args = CMDParser.jepSysArgs()
        (options, args) = parser.parse_args(args)
        if len(args) != 0:
            parser.error("incorrect number of arguments")

        full_build = bool(options.full_build)
        start_dir = CMDParser.jepArgPath(options.start_dir)
        timeout = float(options.timeout)
        allow_include_missing = bool(options.allow_include_missing)
        build_type = options.build_type

        print('FullBuild: {}'.format(full_build))
        print('BuildType: {}'.format(build_type))
        print('Start Path: {}'.format(start_dir))
        print('Single Build Timeout: {:.1f} seconds'.format(timeout))
        print('Include Missing Component Dependencies: {}'.format(allow_include_missing))

        builder = cls(start_dir, full_build, timeout)
        builder.allow_include_missing = allow_include_missing
        builder.stable_build = True if build_type.lower() == 'stable' else False

        return builder

    def deployReports(self):
        """copy unittest files to reports dir"""

        if os.path.isdir(self.build_reports_dir):
            print('Deleting Build Reports Directory: {}'.format(self.build_reports_dir))
            shutil.rmtree(self.build_reports_dir, onerror=fu.on_rm_error)

        os.makedirs(self.build_reports_dir, exist_ok=True)

        print("Moving reports to {}".format(self.build_reports_dir))
        shutil.move(self.junit_build.junit_xml_file, self.build_reports_dir)
        if os.path.isfile(self.junit_changed.junit_xml_file):
            shutil.move(self.junit_changed.junit_xml_file, self.build_reports_dir)

    def startBuild(self):
        """
        Run the build of components
        """
        if self.full_build:
            self.emptyUserGroups()
            print('\nStarting full build of components...\n')
        else:
            print('\nStarting build for new components only...\n')

        print('Start Directory: {}'.format(self.start_path))

        try:
            if os.path.isdir(ComponentHelp.log_dir):
                print('Removing Sphinx Log Directory: {}'.format(ComponentHelp.log_dir))
                shutil.rmtree(ComponentHelp.log_dir, onerror=fu.on_rm_error)
        except Exception as ex:
            print(ex)

        components_list = self.getChangedComponents(write_xml=True)
        sorted_components_list = self.sortComponentslist(components_list)
        if self.full_build:
            keys = set(self.json_report.build.keys()).difference(components_list)
            for key in keys:
                self.json_report.removeEntry(key)
        self.buildComponents(sorted_components_list)

    def getChangedComponents(self, write_xml: bool = False, all_components=False) -> list:
        """
        Get paths of components that has been modified since last build

        Parameters
        ----------

        write_xml : bool
            Write a junit xml, Default (False)

        Returns
        -------

        components_list : list
            List of changed Components

        """

        components_list = []
        self.junit_changed.startTimer()

        all_scripts = 'build_component.py'
        if all_components:
            all_scripts = '*build_component.py'

        file_paths = fu.searchBuildComponentScripts(self.start_path, file_name=all_scripts)
        n_components = len(list(file_paths))
        print('\nFound {} components'.format(n_components))

        for file_path in file_paths:
            t1 = time.time()
            component_path = os.path.dirname(file_path)
            build, error_msg = self.findComponent(component_path)
            if build:
                components_list.append(component_path)
            if error_msg != '':
                dt = time.time() - t1
                self.junit_changed.addTestCase(component_path, dt, error_msg)
        if write_xml:
            self.junit_changed.writeXML()

        if all_components:
            components_list = list(set(components_list))

        return components_list

    def sortComponentslist(self, components_list: list) -> list:
        """
        Try sorting the list of components, in order to build dependencies first.

        Example
        -------
        components_list = [..., Steelstacking, Stator Interior1, WindingTurnConfig1, Winding, ...]
        But 'Stator Interior1' depends on 'Steelstacking', 'WindingTurnConfig1', 'MagnetWireRound', 'Winding'
        Then change the component order so that 'Stator Interior1' appears after its dependencies:
        sorted_components_list = [..., Steelstacking, WindingTurnConfig1, Winding, ..., Stator Interior1, ...]

        Parameters
        ----------
        components_list : list
            List of changed components

        Returns
        -------
        sorted_components_list : list
            List of sorted build order for components
        """

        def toAbsPath(rel_val: str) -> str:
            """replace placeholders for correct absolute path"""
            return rel_val.replace('%ComponentSpace%', Paths.getComponentSpaceRoot())

        if self.allow_include_missing and self.start_path != Paths.getComponentSpaceRoot():
            include_missing = True
        else:
            include_missing = False

        dependency_file = os.path.join(Paths.getComponentSpaceRoot(), 'build_dependencies.json')

        if os.path.isfile(dependency_file):
            '''check if file exists'''

            rel_dependencies = json.loads(open(dependency_file).read())
            abs_dependencies = {}

            for rel_key in rel_dependencies.keys():
                '''convert all relative to absolute paths'''
                abs_key = toAbsPath(rel_key)
                abs_dependencies[abs_key] = []
                for rel_dep in rel_dependencies[rel_key]:
                    abs_dep = toAbsPath(rel_dep)
                    abs_dependencies[abs_key].append(abs_dep)

            enable_sort = False
            sort_dependencies = {}

            print('\nResolving component dependencies')
            for comp_path in components_list:
                try:
                    sort_dependencies[comp_path] = []
                    if comp_path in abs_dependencies:
                        for dep_path in abs_dependencies[comp_path]:
                            if dep_path in components_list:
                                '''enable sorting'''
                                sort_dependencies[comp_path].append(dep_path)
                                enable_sort = True
                            elif include_missing:
                                '''add the dependency and enable sorting'''
                                # Todo is there a need to add missing components recursive
                                #  (for inclusion of dependencies of missing components)?
                                sort_dependencies[comp_path].append(dep_path)
                                if dep_path in abs_dependencies:
                                    sort_dependencies[dep_path] = abs_dependencies[dep_path]
                                else:
                                    sort_dependencies[dep_path] = []
                                enable_sort = True
                except Exception as ex:
                    print('Skipped unexpected error: {} {}'.format(comp_path, str(ex)))

            if enable_sort:
                print('\nSort component build-order')
                sorted_components_list = kahn_topsort(sort_dependencies)
                n_components = len(sorted_components_list)
                for index, path_comp in enumerate(sorted_components_list):
                    disp_path = path_comp.replace(Paths.getComponentSpaceRoot(), '.')
                    print('{}/{} {}'.format(index + 1, n_components, disp_path))
            else:
                sorted_components_list = components_list

        else:
            sorted_components_list = components_list
        return sorted_components_list

    def writeJSONReport(self) -> str:
        """
        Find all buildable components and write them to a json file

        Returns
        -------
        str
            path to json file
        """
        components_list = self.getChangedComponents()

        for index, path_comp in enumerate(components_list):
            revision = ComponentSVNUtils.readRevisionFile(path_comp)
            self.json_report.addRevision(path_comp, revision)
            self.json_report.addVersion(path_comp)
            self.json_report.addPackages(path_comp)

        self.json_report.writeJSON()
        return self.json_report.build_info_json

    def buildComponents(self, components_list: list):
        """
        Try to build all components in the components_list


        Successfully built components are written to a json file

        Parameters
        ----------
        components_list : list
            List of components to build

        """

        n_components = len(components_list)
        print('\nBuilding {} components'.format(n_components))

        self.junit_build.startTimer()

        for index, path_comp in enumerate(components_list):
            revision = ComponentSVNUtils.readRevisionFile(path_comp)

            disp_path = path_comp.replace(Paths.getComponentSpaceRoot(), '.')
            timestamp = str(datetime.datetime.now().strftime('%H:%M:%S'))
            print('{} Build {}/{}: {}'.format(timestamp, index + 1, n_components, disp_path))

            t1 = time.time()
            msg = self.executeBuildScript(path_comp)
            dt = time.time() - t1

            failed = self.junit_build.addTestCase(path_comp, dt, msg)

            if failed:
                self.json_report.removeEntry(path_comp)
                print('        ==> Failed')
            else:
                self.json_report.addRevision(path_comp, revision)
                self.json_report.addVersion(path_comp)
                self.json_report.addPackages(path_comp)
                self.json_report.addBuildTests(path_comp)

            # '''write files for every index'''
            # self.json_report.writeJSON()
            # self.junit_build.writeXML()

        '''write files at the end'''
        self.junit_build.writeTxt()
        self.junit_build.writeXML()
        self.json_report.writeJSON()

    def findComponent(self, component_path: str) -> tuple:
        """
        Try to find a component in the given path

        Parameters
        ----------
        component_path : str
            Path to the component

        Returns
        -------
        tuple
        """
        build = False
        error_message = ''
        disp_path = component_path.replace(Paths.getComponentSpaceRoot(), '.')
        min_version = '1.0' if self.stable_build else '0.0'

        try:
            svn_revision, error = ComponentSVNUtils.writeRevisionFile(component_path)

            if error == '':
                '''get data from previous build'''
                build_revision = self.json_report.getRevision(component_path)
                build_packages = self.json_report.getPackages(component_path)
                build_version = self.json_report.getVersion(component_path)

                has_new_revision = build_revision < svn_revision
                has_new_packages = bool(build_packages)

                if (build_version == '0.0' or has_new_revision) and self.stable_build:
                    '''retrieving the version is slow -> only run if necessary'''
                    JSONReport.writeVersion(component_path)
                    version = JSONReport.readVersion(component_path)
                else:
                    version = build_version

                if self.stable_build:
                    has_min_version = VersionUtil.parse_version(min_version) <= VersionUtil.parse_version(version)
                else:
                    has_min_version = True

                full_version = '{}.{}'.format(version, svn_revision)

                if not has_min_version:
                    build = False
                elif has_new_revision or self.full_build:
                    '''svn version is higher the one from the previous build'''
                    print('Build ==> {}: {}'.format(disp_path, full_version))
                    build = True
                elif has_new_packages:
                    '''loop over all packages if packages is not empty'''
                    for pkg_name, pkg_version in build_packages.items():
                        if pkg_name not in self.pkg_versions:
                            '''package name is not in the list of available packages'''
                            print('Error ==x {} : Unknown package {}'.format(disp_path, pkg_name))
                            error_message = 'Package Error: {}'.format(error)
                            break
                        elif VersionUtil.parse_version(self.pkg_versions[pkg_name]) < VersionUtil.parse_version(pkg_version):
                            '''package version is higher than the previous build'''
                            build = True
                            print('Build ==> {}: {}'.format(disp_path, full_version))
                            print('      ==> {}: {}'.format(pkg_name, self.pkg_versions[pkg_name]))
                            break
                if not build:
                    '''do not build the component'''
                    print('Skip ==> {}: {}'.format(disp_path, full_version))
                    if not has_min_version:
                        print('     ==> required minimum version: {}'.format(min_version))

            else:
                print('Error ==x {} : {}'.format(disp_path, svn_revision))
                error_message = 'SVN Error: {}'.format(error)
        except Exception as e:
            print('Error ==x {}'.format(disp_path))
            error_message = 'Exception Error: {}'.format(e)

        return build, error_message

    def executeBuildScript(self, component_path: str) -> str:
        """
        Execute build script and save error log

        Parameters
        ----------
        component_path : str
            Path to the component
        """
        try:
            build_component_py = pathjoin(component_path, 'build_component.py')
            build_txt = ComponentPaths.getBuildFile(component_path)

            script_py = run_external_buildscript.__file__
            cmd_args = [build_component_py, build_txt]

            # Todo: most reading and writing to build_txt might be unnecessary
            try:
                pu.csymspaceRun(script_py, cmd_args=cmd_args, timeout=self.timeout)
            except TimeoutError as ext:
                with open(build_txt, 'w') as fh:
                    fh.write('Error: TimeoutExpired {}'.format(ext))
            except Exception as ex:
                with open(build_txt, 'w') as fh:
                    fh.write('Error: {}'.format(ex))

            with open(build_txt) as fh:
                msg = fh.read()
        except Exception as ex:
            '''catch all other exception which may occur, e.g. if there is a problem with writing build_txt'''
            msg = 'Error: {}'.format(ex)

        return msg

    @staticmethod
    def emptyUserGroups():
        """
        Remove files from component directories in UserGroups

        """
        print('Emptying _UserGroups ...')
        for gr in os.listdir(pathjoin(Paths.getComponentSpaceRoot(), '_UserGroups')):
            path_comp = pathjoin(Paths.getComponentSpaceRoot(), '_UserGroups', gr, 'Components')
            if os.path.exists(path_comp):
                for fh in os.listdir(path_comp):
                    print('Remove ' + fh)
                    file = pathjoin(path_comp, fh)
                    if os.path.isfile(file):
                        os.remove(file)
                    elif os.path.exists(file):
                        shutil.rmtree(file)


if __name__ == '__main__':
    '''Quickstart for Debugging / Developing'''
    _sp = os.path.join(Paths.getComponentSpaceRoot(), 'Simulation', 'Motor', 'Magnetic', 'PMSM_RBF')
    os.chdir(_sp)
    _builder = ComponentBuilder(_sp, True)
    _builder.startBuild()
