#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.5

$LastChangedRevision: 5103 $
$LastChangedDate:: 2021-10-22 15:30:42 +0200#$

This file is part of SyMSpace.

Copyright (c) 2017, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
from componentspace.base_component import Component
from symspace.wrapper.symwrapper import SymContainer, SymProject


class MaterialComponent(Component):
    """
    Generation of material components
    """
    units = {
        'mm': 'mm',
        'kg.dm-3': 'kg/dm^{3}',
        'kg.m-3': 'kg/m^{3}',
        'N.mm-2': 'N/mm^2',
        'kN.mm-2': 'kN/mm^2',
        'J.kg-1.K-1': r'J/\left(kg\,K\right)',
        'W.m-1.K-1': r'W/\left(m\,K\right)',
        'Ohm.mm2.m-1': r'\Omega\,mm^{2}/m',
        'S.m.mm-2': r'S\,m/mm^{2}',
        'W.kg-1': 'W/kg',
        'K-1': '1/K',
        'degC': r'^{\circ} C',
        'm.V-1': 'm/V',
        'm2.N-1': 'm^2/N',
        'N.V-1.m-1': r'N/\left(V\,m\right)',
        'N.m-2': 'N/m^2',
        'T': 'T'
    }

    def __init__(self, comp_file):
        super().__init__(comp_file)
        self.__base_directory = self.path

    @property
    def base_directory(self):
        return self.__base_directory

    @base_directory.setter
    def base_directory(self, val: str):
        self.__base_directory = val

    def preBuildTest(self) -> dict:
        """
        Get parameters for test file

        Returns
        -------
        d : dict
        """
        mop_file = os.path.join(os.path.dirname(self.path), self.path, self.name)

        if not os.path.isfile(mop_file) and not os.path.isfile(mop_file + ".mop"):
            raise FileNotFoundError(mop_file + " can not be found.")
        sym_mdl = SymProject.load(mop_file)
        if not isinstance(sym_mdl, SymProject):
            raise Exception(mop_file + " can not be loaded.")

        webpath = sym_mdl['Info']['WebPath'].getData()
        comp_webpath = '.'.join([webpath, self.name])

        testparam = sym_mdl['Testparameter']

        sym_mdl.close()

        if testparam is None:
            raise Exception('Container Testparameter does not exist!')

        test_mop_file = os.path.join(self.base_directory, 'tests', 'test_{}'.format(self.name))

        d = {'MopFile': test_mop_file,
             'WebPath': comp_webpath,
             'TestParam': testparam}
        return d

    def generateHelpFiles(self, comp_cont: SymContainer):
        """
        Generate help files

        Parameters
        ----------
        comp_cont : SymContainer
        """
        # Component parameters
        super().generateHelpFiles(comp_cont)

        with open(os.path.join(self.path, 'rst', 'Property.txt'), 'w') as fh:
            fh.write('.. csv-table:: Physical Properties\n' +
                     '    :header: "Property", "", "Value", "Unit", "at Temperature"\n' +
                     '    :widths: 10, 5, 5, 7, 5\n\n')
            for param in comp_cont.getParam():
                if param.getComment().strip().startswith('#') and not isinstance(param, SymContainer):
                    tex = param.getTexString()
                    if len(tex) > 0:
                        tex = ':math:`{}`'.format(tex)

                    if param.getUnit() in self.units:
                        unit = ':math:`{}`'.format(self.units[param.getUnit()])
                    else:
                        unit = param.getUnit()
                    fh.write('    "{}", "{}", "{}", "{}", "{}"\n'.format(
                        param.getDescription(),
                        tex,
                        param.getData(),
                        unit,
                        ''
                    ))
